-- 输出到控制台，调试用
function printf(...)
    ConsoleAddMessage(string.format(...))
end

-- 输出到控制台，调试用
setprinthandler(
    function(...)
        s_args = {}
        for _, arg in ipairs({tostringall(...)}) do
            table.insert(s_args, arg)
        end
        ConsoleAddMessage(table.concat(s_args, ", "))
    end
)

-- 将错误弹窗信息重定向到控制台
seterrorhandler(
    function(err)
        ConsoleAddMessage(err)
        ConsoleAddMessage(debugstack(2))
    end
)

-- 在聊天窗口打印信息
function print_chat(...)
    s_args = {}
    for _, arg in ipairs({tostringall(...)}) do
        table.insert(s_args, arg)
    end
    DEFAULT_CHAT_FRAME:AddMessage(table.concat(s_args, ", "))
end

-- 设置位置
function set_point(f, p, r, rp, x, y)
    f:ClearAllPoints()
    f:SetPoint(p, r, rp, x, y)
end

-- 获取包含所有的大小
function get_size(...)
    local top, left, right, bottom = {}, {}, {}, {}
    for _, region in ipairs({...}) do
        table.insert(top, region:GetTop())
        table.insert(left, region:GetLeft())
        table.insert(right, region:GetRight())
        table.insert(bottom, region:GetBottom())
    end
    return math.max(unpack(right)) - math.min(unpack(left)), math.max(unpack(top)) - math.min(unpack(bottom))
end

function get_table_size(t)
    local count = 0
    for _, __ in pairs(t) do
        count = count + 1
    end
    return count
end

-- 接收事件的匿名框架
local callbacks = {}
local frame = CreateFrame("Frame")
frame:Hide()
frame:SetScript(
    "OnEvent",
    function(self, event, ...)
        -- printf("OnEvent: %s", event)
        cbs = callbacks[event]
        if cbs then
            for cb, _ in pairs(cbs) do
                pcall(cb, ...)
            end
        end
    end
)

-- 注册一个事件回调函数
function register_event(e, cb)
    -- print("register_event", e, cb)
    if not callbacks[e] then
        callbacks[e] = {}
        frame:RegisterEvent(e)
    end
    callbacks[e][cb] = true
end

-- 注销一个事件回调函数
function unregister_event(e, cb)
    -- print("unregister_event", e, cb)
    if callbacks[e] then
        callbacks[e][cb] = nil
        if get_table_size(callbacks[e]) == 0 then
            callbacks[e] = nil
            frame:UnregisterEvent(e)
        end
    end
end

-- 创建一个响应点击事件的按钮
function create_button(text, onclick)
    local btn = CreateFrame("Button", nil, UIParent, "UIPanelButtonTemplate")
    btn:SetText(text)
    btn:SetSize(btn:GetTextWidth() + 16, btn:GetTextHeight() + 8)
    btn:SetScript("OnClick", onclick)
    return btn
end

function hook_non_recursive(t, fn, cb)
    local is_in_hook = false -- 防止无限循环调用
    hooksecurefunc(
        t,
        fn,
        function(...)
            if not is_in_hook then
                is_in_hook = true
                pcall(cb, ...)
                is_in_hook = false
            end
        end
    )
end

-- 通过hook SetPoint函数强制重新调整位置
function hook_set_point(t, cb)
    -- 第一次调用
    pcall(cb) -- 有可能失败
    hook_non_recursive(t, "SetPoint", cb)
end

function starts_with(str, start)
    return str:sub(1, #start) == start
end

function ends_with(str, ending)
    return ending == "" or str:sub(-(#ending)) == ending
end
