nearby_players_list_position = nearby_players_list_position or {0, 0}

local player_list_anchor = nil
local unused_player_buttons = {}
local visible_player_buttons = {}
local last_seen = {}
local alliance_color = {{0.6, 0.6, 1, 1}, {0, 0, 0.4, 0.6}}
local horde_color = {{1, 0.6, 0.6, 1}, {0.4, 0, 0, 0.6}}
local neutral_color = {{0.8, 0.8, 0.8, 1}, {0.2, 0.2, 0.2, 0.6}}
local race_color = {
    -- https://wow.gamepedia.com/API_UnitRace
    Dwarf = alliance_color,
    DarkIronDwarf = alliance_color,
    Draenei = alliance_color,
    Gnome = alliance_color,
    Human = alliance_color,
    LightforgedDraenei = alliance_color,
    NightElf = alliance_color,
    Pandaren = alliance_color,
    VoidElf = alliance_color,
    Worgen = alliance_color,
    KulTiran = alliance_color,
    BloodElf = horde_color,
    Goblin = horde_color,
    HighmountainTauren = horde_color,
    MagharOrc = horde_color,
    Orc = horde_color,
    Pandaren = horde_color,
    Tauren = horde_color,
    Troll = horde_color,
    Scourge = horde_color,
    Nightborne = horde_color,
    ZandalariTroll = horde_color,
    Pandaren = neutral_color
}

local function player_button_on_enter(self)
    GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
    GameTooltip:ClearLines()
    GameTooltip:AddDoubleLine(self.tooltip_text_1, self.tooltip_text_2, 1, 1, 1, unpack(self.tooltip_color))
    GameTooltip:Show()
end

local function player_button_on_leave(self)
    GameTooltip:Hide()
end

local function add_or_refresh_player(guid)
    if InCombatLockdown() then
        return
    end
    local inst, _ = IsInInstance()
    if inst then
        return
    end
    if not guid:find("^Player") then
        return
    end
    -- print(guid, name)
    if last_seen[guid] == nil then
        -- 如果不存在则新增
        local btn = table.remove(unused_player_buttons)
        if btn == nil then
            -- print("创建frame")
            -- 固定属性
            btn = CreateFrame("Button", nil, UIParent, "SecureActionButtonTemplate")
            btn:SetAttribute("type", "macro")
            btn:SetNormalFontObject("GameFontNormal")
            btn:SetNormalTexture(btn:CreateTexture(nil))
            btn:GetNormalTexture():SetAllPoints(true)
            btn:SetHighlightTexture(btn:CreateTexture(nil))
            btn:GetHighlightTexture():SetAllPoints(true)
            btn:GetHighlightTexture():SetColorTexture(1, 1, 1, 0.1)
            btn:SetScript("OnEnter", player_button_on_enter)
            btn:SetScript("OnLeave", player_button_on_leave)
            btn:SetScale(1)
        end
        -- 可变属性
        local cn, cid, rn, rid, gender, name, realm = GetPlayerInfoByGUID(guid)
        btn.player_guid = guid
        btn.tooltip_text_1 = name
        btn.tooltip_text_2 = rn .. cn
        btn.tooltip_color = {GetClassColor(cid)}
        btn:SetText(name)
        btn:SetAttribute("macrotext", "/target " .. name)
        btn:SetSize(btn:GetTextWidth() + 12, btn:GetTextHeight() + 4)
        local rcolor = race_color[rid]
        if not rcolor then
            print_chat(rid)
            rcolor = neutral_color
        end
        btn:GetFontString():SetTextColor(rcolor[1][1], rcolor[1][2], rcolor[1][3], rcolor[1][4])
        btn:GetNormalTexture():SetColorTexture(rcolor[2][1], rcolor[2][2], rcolor[2][3], rcolor[2][4])
        -- 添加到界面上
        table.insert(visible_player_buttons, btn)
        btn:ClearAllPoints()
        btn:Show()
        btn:SetPoint("TOPLEFT", player_list_anchor, "BOTTOMLEFT", 0, 0)
        if #visible_player_buttons > 1 then
            visible_player_buttons[#visible_player_buttons - 1]:SetPoint("TOPLEFT", btn, "BOTTOMLEFT", 0, 0)
        end
    -- print("新增" .. guid .. " " .. #visible_player_buttons .. " " .. #unused_player_buttons)
    end
    -- 刷新时间
    last_seen[guid] = time()
end

local function remove_player(guid)
    if InCombatLockdown() then
        return
    end
    last_seen[guid] = nil
    -- 查找并删除
    local idx, btn = nil, nil
    for i, b in ipairs(visible_player_buttons) do
        if b.player_guid == guid then
            idx, btn = i, b
            break
        end
    end
    if idx then
        -- 重新调整位置
        local prev_frame = visible_player_buttons[idx - 1]
        local next_frame = visible_player_buttons[idx + 1]
        if prev_frame then
            prev_frame:ClearAllPoints()
            if next_frame then
                prev_frame:SetPoint("TOPLEFT", next_frame, "BOTTOMLEFT", 0, 0)
            else
                prev_frame:SetPoint("TOPLEFT", player_list_anchor, "BOTTOMLEFT", 0, 0)
            end
        end
        -- 删除
        btn:ClearAllPoints()
        btn:Hide()
        table.remove(visible_player_buttons, idx)
        table.insert(unused_player_buttons, btn)
    -- print("删除" .. btn.player_guid .. " " .. #visible_player_buttons .. " " .. #unused_player_buttons)
    end
end

local function remove_disappeared_players()
    if InCombatLockdown() then
        return
    end
    local now = time()
    -- 移除长时间未见的
    local remove_list = {}
    for id, time in pairs(last_seen) do
        if (now - time) > 60 then
            table.insert(remove_list, id)
        end
    end
    for _, id in ipairs(remove_list) do
        remove_player(id)
    end
end

function setup_player_detector()
    player_list_anchor = CreateFrame("Frame", nil, UIParent)
    player_list_anchor:SetMovable(true)
    player_list_anchor:EnableMouse(true)
    player_list_anchor:RegisterForDrag("LeftButton")
    player_list_anchor:SetScript("OnDragStart", player_list_anchor.StartMoving)
    player_list_anchor:SetScript(
        "OnDragStop",
        function(this)
            _, _, _, nearby_players_list_position[0], nearby_players_list_position[1] = this:GetPoint()
            return this:StopMovingOrSizing()
        end
    )
    player_list_anchor:SetPoint(
        "TOPLEFT",
        UIParent,
        "TOPLEFT",
        nearby_players_list_position[0],
        nearby_players_list_position[1]
    )
    local tex = player_list_anchor:CreateTexture("ARTWORK")
    tex:SetAllPoints()
    tex:SetColorTexture(0.4, 0.4, 0.4, 0.6)
    local fo = player_list_anchor:CreateFontString(nil)
    fo:SetFontObject("GameFontNormal")
    fo:SetTextColor(1, 1, 1, 1)
    fo:SetAllPoints()
    fo:SetText("≡ 附近玩家 ≡")
    player_list_anchor:SetSize(fo:GetStringWidth() + 12, fo:GetStringHeight() + 4)

    -- 战斗日志
    register_event(
        "COMBAT_LOG_EVENT_UNFILTERED",
        function()
            local _, _, _, sguid, sname, _, _, dguid, dname, _, _ = CombatLogGetCurrentEventInfo()
            add_or_refresh_player(sguid)
            add_or_refresh_player(dguid)
        end
    )

    -- 聊天
    local function chat_detector(...)
        local arg = {...}
        local guid = arg[12]
        add_or_refresh_player(guid)
    end
    register_event("CHAT_MSG_SAY", chat_detector)
    register_event("CHAT_MSG_YELL", chat_detector)
    -- register_event("CHAT_MSG_WHISPER", chat_detector)
    register_event("CHAT_MSG_TEXT_EMOTE", chat_detector)
    register_event("CHAT_MSG_EMOTE", chat_detector)
    register_event("CHAT_MSG_FILTERED", print_chat)

    -- 鼠标点选或划过
    register_event(
        "PLAYER_TARGET_CHANGED",
        function()
            add_or_refresh_player(UnitGUID("target"))
        end
    )
    register_event(
        "UPDATE_MOUSEOVER_UNIT",
        function()
            add_or_refresh_player(UnitGUID("mouseover"))
        end
    )

    -- 自动清理
    C_Timer.NewTicker(
        30,
        function()
            -- print_chat("Cleanup")
            remove_disappeared_players()
        end
    )
end
