local N, T = ...

function T.Setup_Combatlog()
    combat_statistics = combat_statistics or {}
    combat_statistics.damage_heal = combat_statistics.damage_heal or {} -- { GUID = {伤害，治疗}}
    combat_statistics.minion_owner = combat_statistics.minion_owner or {} -- {宠物召唤物等的GUID = 拥有者的GUID}
    combat_statistics.id_name = combat_statistics.id_name or {} -- {GUID = 名字}，注意名字不要用格式字符串，否则Invalid escape code in chat message错误或者服务器不理睬
    combat_statistics.duration = combat_statistics.duration or 0 -- 总战斗持续时间

    local my_realm = GetRealmName()

    local combat_start_time, combat_end_time = 0, 0

    local function record_interested(uid)
        -- 通过unitid查询玩家及其pet（如果有的话）的id、名字以及从属关系
        -- 将只记录id_names中存在的战斗数据
        local unit_guid = UnitGUID(uid)
        local unit_name, unit_realm = UnitName(uid)
        -- print(unit_guid, unit_name, unit_realm)
        if unit_guid then
            if unit_realm and unit_realm ~= "" then -- 非本服务器角色名补全realm name
                unit_name = unit_name .. "-" .. unit_realm
            end
            combat_statistics.id_name[unit_guid] = unit_name
            local pid = uid .. "pet"
            local pet_guid, pet_name = UnitGUID(pid), UnitName(pid)
            -- print(pet_guid, pet_name)
            if pet_guid then
                combat_statistics.minion_owner[pet_guid] = unit_guid
            end
        end
    end

    -- 进入游戏时候记录自己的信息
    record_interested("player")

    local function reset()
        combat_statistics = {}
        combat_statistics.damage_heal = {}
        combat_statistics.minion_owner = {}
        combat_statistics.id_name = {}
        combat_statistics.duration = 0
        -- 该程序感兴趣的包括玩家以及组队/团队下的队友
        local uid
        if IsInRaid() then
            for i = 1, MAX_RAID_MEMBERS do
                uid = "raid" .. i
                record_interested(uid)
            end
        elseif IsInGroup() then
            for i = 1, MAX_PARTY_MEMBERS do
                uid = "party" .. i
                record_interested(uid)
            end
            record_interested("player") -- MAX_PARTY_MEMBERS=4，不包括自己
        else
            record_interested("player")
        end
        print("Damage Meter Reset")
    end

    local function summary()
        return coroutine.wrap(
            function()
                local t = {}
                local name
                for id, sum in pairs(combat_statistics.damage_heal) do
                    name = combat_statistics.id_name[id]
                    if name then
                        if id:find("^Player-") and (not name:find("-")) then
                            name = name .. "-" .. my_realm -- 角色名统统补全realm name
                        end
                    else
                        name = id
                    end
                    table.insert(t, {name, sum[1], sum[2]})
                end
                local duration = max(combat_end_time - combat_start_time + combat_statistics.duration, 1) -- 至少1秒
                local function isint(n)
                    return n % 1 == 0
                end
                local function round(n)
                    return n > 0 and math.floor(n + 0.5) or math.ceil(n - 0.5)
                end
                local function numf(n)
                    if type(n) ~= "number" then
                        return tostring(n)
                    elseif n < 10000 then
                        return string.format("%d", n)
                    elseif n < 100000 then
                        return string.format("%.1fk", n / 1000)
                    elseif n < 1000000000 then
                        return string.format("%.1fw", n / 10000)
                    else
                        return string.format("%.1fe", n / 100000000)
                    end
                end
                table.sort(
                    t,
                    function(e1, e2)
                        return e1[2] > e2[2]
                    end
                )
                coroutine.yield("伤害:")
                for i, sum in ipairs(t) do
                    coroutine.yield(string.format("%d. %s %s (%s)", i, sum[1], numf(sum[2]), numf(sum[2] / duration)))
                end
                -- table.sort(
                --     t,
                --     function(e1, e2)
                --         return e1[3] > e2[3]
                --     end
                -- )
                -- coroutine.yield("治疗:")
                -- for i, sum in ipairs(t) do
                --     coroutine.yield(string.format("%d. %s %s (%s)", i, sum[1], numf(sum[3]), numf(sum[3] / duration)))
                -- end
            end
        )
    end

    local function on_combat_log(...)
        -- print(...)
        local timestamp, event, _, source_guid, source_name, _, _, dest_guid, dest_name, _, _ = ...

        local spell_id, spell_name, spell_school, environmental_type
        local suff
        if event:find("^SWING") then
            suff = 12
        elseif event:find("^ENVIRONMENTAL") then
            environmental_type = select(12, ...)
            suff = 13
        else
            spell_id, spell_name, spell_school = select(12, ...)
            suff = 15
        end

        local function get_owner(id)
            local mid = id
            local oid = combat_statistics.minion_owner[mid]
            while oid do
                mid = oid
                oid = combat_statistics.minion_owner[mid]
            end
            return mid
        end
        local function add_damage_heal(id, damage, heal)
            local oid = get_owner(id)
            if not combat_statistics.id_name[oid] then
                return
            end
            if not combat_statistics.damage_heal[oid] then
                combat_statistics.damage_heal[oid] = {0, 0}
            end
            combat_statistics.damage_heal[oid][1] = combat_statistics.damage_heal[oid][1] + damage
            combat_statistics.damage_heal[oid][2] = combat_statistics.damage_heal[oid][2] + heal
        end
        local amount, overkill, overhealing
        if event:find("_DAMAGE$") then
            if InCombatLockdown() then -- 只记录战斗状态下的
                amount, overkill = select(suff, ...)
                add_damage_heal(source_guid, amount, 0)
                combat_end_time = timestamp
                -- print(timestamp, time())
            end
        elseif event:find("_HEAL$") then
            if InCombatLockdown() then -- 只记录战斗状态下的
                amount, overhealing = select(suff, ...)
                add_damage_heal(source_guid, 0, amount)
                combat_end_time = timestamp
            end
        elseif event:find("_SUMMON$") then
            if combat_statistics.id_name[source_guid] then
                combat_statistics.minion_owner[dest_guid] = source_guid -- 战斗非战斗状态都应当记录
            end
        end
    end

    SLASH_COMBAT_LOG1 = "/cl"
    SlashCmdList["COMBAT_LOG"] = function(msg)
        if msg == "" then
            for line in summary() do
                print(line)
            end
        elseif msg == "say" then
            local chat_type
            if IsInRaid(LE_PARTY_CATEGORY_INSTANCE) then
                chat_type = "RAID"
            elseif IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
                chat_type = "INSTANCE_CHAT"
            else
                chat_type = "EMOTE"
            end
            for line in summary() do
                -- line = line:gsub("|", "\124")
                SendChatMessage(line, chat_type)
            end
        elseif msg == "rst" then
            reset()
        end
    end

    local frame = CreateFrame("Frame")
    frame:Hide()
    frame:SetScript(
        "OnEvent",
        function(self, event, ...)
            if event == "COMBAT_LOG_EVENT_UNFILTERED" then
                on_combat_log(CombatLogGetCurrentEventInfo())
            elseif event == "PLAYER_ENTERING_WORLD" then
                -- print(...)
                local is_initial_login, is_reloading_ui = ...
                if not (is_initial_login or is_reloading_ui) then
                    -- 进入副本重置信息
                    -- 要塞也是副本，从副本退回到要塞会重置
                    name, instance_type = GetInstanceInfo()
                    -- print(name, instance_type)
                    if instance_type ~= "none" and (not name:find('Garrison')) then -- 德拉诺要塞需要排除
                        reset()
                    end
                end
            elseif event == "UNIT_PET" then
                local uid = ...
                record_interested(uid)
            elseif event == "PLAYER_REGEN_DISABLED" then
                combat_start_time = time()
                combat_end_time = combat_start_time
            elseif event == "PLAYER_REGEN_ENABLED" then
                combat_statistics.duration = combat_statistics.duration + (combat_end_time - combat_start_time)
                combat_start_time = combat_end_time
            end
        end
    )
    frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    frame:RegisterEvent("PLAYER_ENTERING_WORLD")
    frame:RegisterEvent("UNIT_PET")
    frame:RegisterEvent("PLAYER_REGEN_DISABLED")
    frame:RegisterEvent("PLAYER_REGEN_ENABLED")
end
