local N, T = ...

function T.Setup_Item()
    -- Interface\FrameXML\ContainerFrame.lua
    -- Interface\FrameXML\BankFrame.lua
    local function toggle_bags(from, to)
        assert(from >= 0)
        assert(to <= 12)
        local bags_available = {}
        local bags_opened = {}
        for i = 0, 12 do
            if GetContainerNumSlots(i) > 0 then -- 银行窗口关闭时候银行的包也返回0
                bags_available[i] = true
                if IsBagOpen(i) then
                    bags_opened[i] = true
                    CloseBag(i) -- 先全部关闭再打开，达到从头排序的效果
                end
            end
        end
        local toggle = false
        for i = from, to do -- 计算到底是开还是关
            toggle = toggle or (bags_available[i] ~= bags_opened[i])
        end
        for i = from, to do
            bags_opened[i] = toggle
        end
        for i = 0, 12 do -- 全部打开
            if bags_opened[i] then
                OpenBag(i)
            end
        end
    end
    ToggleAllBags = function()
        -- 替换全局函数，比如按B键
        toggle_bags(0, 12)
    end
    local function toggle_equipped_bags()
        toggle_bags(0, 4)
    end
    local function toggle_bank_bags()
        toggle_bags(5, 12)
    end
    BackpackButton_OnClick = function(self)
        if (not PutItemInBackpack()) then
            toggle_equipped_bags()
        end
    end
    BagSlotButton_OnClick = function(self)
        if (not PutItemInBag(self:GetID())) then
            toggle_equipped_bags()
        end
    end
    hooksecurefunc(BankFrame, "Show", toggle_bank_bags)
    BankFrameItemButtonBag_OnClick = function(self, button)
        if (not PutItemInBag(self:GetInventorySlot())) then
            toggle_bank_bags()
        end
    end
    for i = 1, NUM_CONTAINER_FRAMES do
        hooksecurefunc(
            _G["ContainerFrame" .. i],
            "Show",
            function(self)
                local bag_id = self:GetID()
                -- print(bag_id)
                _G["ContainerFrame" .. i .. "CloseButton"]:SetEnabled(bag_id < 5 or bag_id > 11)
            end
        )
    end
    for i = 1, NUM_CONTAINER_FRAMES do
        _G["ContainerFrame" .. i .. "CloseButton"]:HookScript(
            "OnClick",
            function(self)
                local bag_id = self:GetParent():GetID()
                if bag_id < 5 then
                    for i = 0, 4 do
                        CloseBag(i)
                    end
                end
            end
        )
    end
    -- 公会仓库打开
    T.On_AddOn_Loaded(
        "Blizzard_GuildBankUI",
        function()
            GuildBankFrame:HookScript(
                "OnShow",
                function()
                    for i = 0, 4 do
                        CloseBag(i)
                    end
                    for i = 0, 4 do
                        OpenBag(i)
                    end
                end
            )
        end
    )

    local overlayed_item_buttons = {} -- 避免全局污染
    local function item_button_prepare(button, more)
        if not overlayed_item_buttons[button] then
            -- learned from SimpleItemLevel
            local overlay = CreateFrame("FRAME", nil, button)
            overlay:SetFrameLevel(4) -- Azerite overlay must be overlaid itself...
            overlay:SetAllPoints()
            local ilvl = overlay:CreateFontString(nil, "OVERLAY")
            ilvl:SetPoint("TOPRIGHT", -2, -2)
            ilvl:SetFont("Fonts\\ARHei.TTF", 14, "OUTLINE")
            ilvl:SetJustifyH("RIGHT")
            ilvl:Hide()
            local ilvlup = overlay:CreateTexture(nil, "OVERLAY")
            ilvlup:SetSize(8, 8)
            ilvlup:SetPoint("TOPLEFT", 2, -2)
            ilvlup:SetAtlas("poi-door-arrow-up") -- MiniMap-PositionArrowUp?
            ilvlup:Hide()
            local mog = overlay:CreateTexture(nil, "OVERLAY")
            mog:SetSize(16, 16)
            mog:SetPoint("BOTTOMRIGHT", -2, 2)
            mog:Hide()
            overlayed_item_buttons[button] = {
                overlay = overlay,
                ilvl = ilvl,
                ilvlup = ilvlup,
                mog = mog
            }
            if more then
                more(overlayed_item_buttons[button])
            end
        end
        return overlayed_item_buttons[button]
    end
    -- local function hide_ilvl(button)
    --     if overlayed_item_buttons[button] then
    --         overlayed_item_buttons[button].ilvl:Hide()
    --         overlayed_item_buttons[button].ilvlup:Hide()
    --     end
    -- end
    -- local function show_ilvl(button, level, quality)
    --     local entry = item_button_prepare(button)
    --     entry.ilvl:SetText(level)
    --     entry.ilvl:SetTextColor(GetItemQualityColor(quality))
    --     entry.ilvl:Show()
    -- end
    -- local function show_ilvlup(button, red)
    --     local entry = item_button_prepare(button)
    --     if red then
    --         entry.ilvlup:SetVertexColor(1, 0, 0)
    --     else
    --         entry.ilvlup:SetVertexColor(1, 1, 1)
    --     end
    --     entry.ilvlup:Show()
    -- end
    -- local function hide_mog(button)
    --     if overlayed_item_buttons[button] then
    --         overlayed_item_buttons[button].mog:Hide()
    --     end
    -- end
    -- local function show_mog(button, can_collect)
    --     local entry = item_button_prepare(button)
    --     if can_collect then
    --         entry.mog:SetTexture("Interface/AddOns/" .. N .. "/res/unknown.tga")
    --     else
    --         entry.mog:SetTexture("Interface/AddOns/" .. N .. "/res/unknowable.tga")
    --     end
    --     entry.mog:Show()
    -- end
    local function item_button_hide_all(button)
        if overlayed_item_buttons[button] then
            overlayed_item_buttons[button].ilvl:Hide()
            overlayed_item_buttons[button].ilvlup:Hide()
            overlayed_item_buttons[button].mog:Hide()
        end
    end
    local function item_button_show_ilvl(button, item)
        -- https://wowpedia.fandom.com/wiki/Enum.InventoryType
        -- 不是GetItemInventoryType()
        local item_inventory_type = item:GetInventoryType()
        if not item_inventory_type or item_inventory_type == 0 or item_inventory_type == 18 then
            return
        end
        local level = item:GetCurrentItemLevel()
        if not level then
            return
        end
        local entry = item_button_prepare(button)
        entry.ilvl:SetText(level)
        entry.ilvl:SetTextColor(1, 1, 1)
        entry.ilvl:Show()
        local quality = item:GetItemQuality()
        if not quality then
            return
        end
        local r, g, b = GetItemQualityColor(quality)
        if not r then
            return
        end
        entry.ilvl:SetTextColor(r, g, b)
    end
    hooksecurefunc(
        "PaperDollItemSlotButton_Update",
        function(button)
            -- hide_ilvl(button)
            -- local slot_id = button:GetID()
            -- if slot_id < 1 or slot_id > 19 then
            --     return
            -- end
            -- -- 此处如果使用Get...Item...()函数的方式会间歇地level, quality取到nil
            -- local item = Item:CreateFromEquipmentSlot(slot_id)
            -- if item:IsItemEmpty() then
            --     return
            -- end
            -- item:ContinueOnItemLoad(
            --     function()
            --         show_ilvl(button, item:GetCurrentItemLevel(), item:GetItemQuality())
            --     end
            -- )
            ----------------------------------
            item_button_hide_all(button)
            local slot_id = button:GetID()
            if slot_id < 1 or slot_id > 19 then
                return
            end
            local item = Item:CreateFromEquipmentSlot(slot_id)
            if not item or item:IsItemEmpty() then
                return
            end
            item:ContinueOnItemLoad(
                function()
                    item_button_show_ilvl(button, item)
                end
            )
        end
    )
    T.On_AddOn_Loaded(
        "Blizzard_InspectUI",
        function()
            hooksecurefunc(
                "InspectPaperDollItemSlotButton_Update",
                function(button)
                    -- hide_ilvl(button)
                    -- local slot_id = button:GetID()
                    -- if slot_id < 1 or slot_id > 19 then
                    --     return
                    -- end
                    -- local item_link = GetInventoryItemLink("target", slot_id)
                    -- if not item_link then
                    --     return
                    -- end
                    -- local quality = GetInventoryItemQuality("target", slot_id)
                    -- local level = GetDetailedItemLevelInfo(item_link)
                    -- show_ilvl(button, level, quality)
                    ---------------------------------------
                    item_button_hide_all(button)
                    local slot_id = button:GetID()
                    if slot_id < 1 or slot_id > 19 then
                        return
                    end
                    local item_link = GetInventoryItemLink(InspectFrame.unit, slot_id)
                    if not item_link then
                        return
                    end
                    local item = Item:CreateFromItemLink(item_link)
                    if not item or item:IsItemEmpty() then
                        return
                    end
                    item:ContinueOnItemLoad(
                        function()
                            item_button_show_ilvl(button, item)
                        end
                    )
                end
            )
        end
    )
    local function item_button_show_ilvlup(button, item)
        -- https://wowpedia.fandom.com/wiki/Enum.InventoryType
        local inv_type = item:GetInventoryType()
        -- print(inv_type)
        local slot_ids = {C_Transmog.GetSlotForInventoryType(inv_type + 1)} -- 注意+1
        if slot_ids[1] == 0 then
            slot_ids =
                ({
                [Enum.InventoryType.IndexNeckType] = {INVSLOT_NECK},
                [Enum.InventoryType.IndexFingerType] = {INVSLOT_FINGER1, INVSLOT_FINGER2},
                [Enum.InventoryType.IndexTrinketType] = {INVSLOT_TRINKET1, INVSLOT_TRINKET2}
            })[inv_type]
        end
        if not slot_ids then
            -- print("不可装备")
            return -- 不可装备
        end
        local item_id = item:GetItemID()
        local item_link = item:GetItemLink()
        local _, source_id = C_TransmogCollection.GetItemInfo(item_id)
        local can_collect = true
        if source_id then
            _, can_collect = C_TransmogCollection.PlayerCanCollectSource(source_id)
        end
        if not can_collect then
            -- print("本职业无法幻化的")
            return -- 本职业无法幻化的
        end
        local item_level = item:GetCurrentItemLevel()
        local min_level_required = item_link and select(5, GetItemInfo(item_link or item_id))
        -- print(item_level, min_level_required)
        local r, g, b = 1, 1, 1
        if min_level_required and min_level_required > UnitLevel("player") then
            r, g, b = 1, 0, 0
        end
        local function show()
            local entry = item_button_prepare(button)
            entry.ilvlup:SetVertexColor(r, g, b)
            entry.ilvlup:Show()
        end
        for _, slot_id in ipairs(slot_ids) do
            local equipped = Item:CreateFromEquipmentSlot(slot_id)
            if equipped:IsItemEmpty() then
                show()
            else
                equipped:ContinueOnItemLoad(
                    function()
                        if equipped:GetCurrentItemLevel() < item_level then
                            show()
                        end
                    end
                )
            end
        end
    end
    local function get_item_mog_tex(item)
        local item_id = item:GetItemID()
        if not C_Transmog.CanTransmogItem(item_id) then
            return nil
        end
        local appearance_id, source_id = C_TransmogCollection.GetItemInfo(item_id)
        if not appearance_id or not source_id then
            return nil
        end
        if C_TransmogCollection.PlayerHasTransmog(item_id) then
            return nil
        end
        if select(2, C_TransmogCollection.PlayerCanCollectSource(source_id)) then
            local known_appearance = false
            for i, v in ipairs(C_TransmogCollection.GetAllAppearanceSources(appearance_id)) do
                if C_TransmogCollection.GetSourceInfo(v).isCollected then
                    known_appearance = true
                    break
                end
            end
            if known_appearance then
                return "Interface/AddOns/" .. N .. "/res/know_appearance.tga"
            else
                return "Interface/AddOns/" .. N .. "/res/unknown.tga"
            end
        else
            return "Interface/AddOns/" .. N .. "/res/unknowable.tga"
        end
    end
    local function item_button_show_mog(button, item)
        local tex = get_item_mog_tex(item)
        if not tex then
            return
        end
        local entry = item_button_prepare(button)
        entry.mog:SetTexture(tex)
        entry.mog:Show()
    end
    local function update_container_item(button)
        -- hide_ilvl(button)
        -- hide_mog(button)
        -- local slot_id = button:GetID()
        -- local bag_id = button:GetParent():GetID()
        -- local loc = ItemLocation:CreateFromBagAndSlot(bag_id, slot_id)
        -- if not loc:IsValid() then
        --     return
        -- end
        -- local item = Item:CreateFromItemLocation(loc)
        -- if item:IsItemEmpty() then
        --     return
        -- end
        -- item:ContinueOnItemLoad(
        --     function()
        --         -- https://wowpedia.fandom.com/wiki/Enum.InventoryType
        --         -- 不是GetItemInventoryType()
        --         local item_inventory_type = item:GetInventoryType()
        --         if not item_inventory_type or item_inventory_type == 0 then
        --             return
        --         end
        --         local item_level = item:GetCurrentItemLevel()
        --         local item_quality = item:GetItemQuality()
        --         show_ilvl(button, item_level, item_quality)
        --         local item_id = item:GetItemID()
        --         local item_link = item:GetItemLink()
        --         local _, source_id = C_TransmogCollection.GetItemInfo(item_id)
        --         local can_collect
        --         if source_id then
        --             _, can_collect = C_TransmogCollection.PlayerCanCollectSource(source_id)
        --         end
        --         if not source_id or can_collect then -- 项链戒指饰品或者可幻化的装备
        --             local slot_ids = {C_Transmog.GetSlotForInventoryType(item_inventory_type + 1)} -- 注意+1
        --             if slot_ids[1] == 0 then
        --                 slot_ids =
        --                     ({
        --                     [Enum.InventoryType.IndexNeckType] = {INVSLOT_NECK},
        --                     [Enum.InventoryType.IndexFingerType] = {INVSLOT_FINGER1, INVSLOT_FINGER2},
        --                     [Enum.InventoryType.IndexTrinketType] = {INVSLOT_TRINKET1, INVSLOT_TRINKET2}
        --                 })[item_inventory_type]
        --             end
        --             if not slot_ids then
        --                 -- 比如背包类型
        --                 return
        --             end
        --             local min_level_required = item_link and select(5, GetItemInfo(item_link or item_id))
        --             local red = min_level_required and min_level_required > UnitLevel("player")
        --             for _, slot_id in ipairs(slot_ids) do
        --                 local equipped = Item:CreateFromEquipmentSlot(slot_id)
        --                 if equipped:IsItemEmpty() then
        --                     show_ilvlup(button, red)
        --                 else
        --                     equipped:ContinueOnItemLoad(
        --                         function()
        --                             if equipped:GetCurrentItemLevel() < item_level then
        --                                 show_ilvlup(button, red)
        --                             end
        --                         end
        --                     )
        --                 end
        --             end
        --         end
        --         if not C_Item.IsBound(loc) and not C_TransmogCollection.PlayerHasTransmog(item_id) and source_id then
        --             show_mog(button, can_collect)
        --         end
        --     end
        -- )
        -----------------------------
        item_button_hide_all(button)
        local slot_id = button:GetID()
        local bag_id = button:GetParent():GetID()
        local loc = ItemLocation:CreateFromBagAndSlot(bag_id, slot_id)
        if not loc:IsValid() then
            return
        end
        local item = Item:CreateFromItemLocation(loc)
        if not item or item:IsItemEmpty() then
            return
        end
        item:ContinueOnItemLoad(
            function()
                item_button_show_ilvl(button, item)
                item_button_show_ilvlup(button, item)
                if not C_Item.IsBound(loc) then
                    item_button_show_mog(button, item)
                end
            end
        )
    end
    hooksecurefunc(
        "ContainerFrame_Update",
        function(container)
            local name = container:GetName()
            for i = 1, container.size, 1 do
                local button = _G[name .. "Item" .. i]
                update_container_item(button)
            end
        end
    )
    -- 银行界面，非银行背包
    hooksecurefunc(
        "BankFrameItemButton_Update",
        function(button)
            if not button.isBag then
                update_container_item(button)
            end
        end
    )
    -- 商人
    hooksecurefunc(
        "MerchantFrame_Update",
        function()
            for i = 1, MERCHANT_ITEMS_PER_PAGE do
                local button = _G["MerchantItem" .. i .. "ItemButton"]
                item_button_hide_all(button)
                if button.hasItem and button.link then
                    local item = Item:CreateFromItemLink(button.link)
                    if item and not item:IsItemEmpty() then
                        item:ContinueOnItemLoad(
                            function()
                                item_button_show_ilvl(button, item)
                                item_button_show_ilvlup(button, item)
                                item_button_show_mog(button, item)
                            end
                        )
                    end
                end
            end
        end
    )
    -- 冒险指南
    T.On_AddOn_Loaded(
        "Blizzard_EncounterJournal",
        function()
            hooksecurefunc(
                "EncounterJournal_SetLootButton",
                function(button)
                    item_button_hide_all(button)
                    if button.link then
                        local item = Item:CreateFromItemLink(button.link)
                        if item and not item:IsItemEmpty() then
                            item:ContinueOnItemLoad(
                                function()
                                    item_button_prepare(
                                        button,
                                        function(entry)
                                            entry.overlay:ClearAllPoints()
                                            entry.overlay:SetAllPoints(button.lootFrame.icon)
                                            entry.overlay:SetFrameLevel(6)
                                        end
                                    )
                                    item_button_show_ilvl(button, item)
                                    item_button_show_ilvlup(button, item)
                                    item_button_show_mog(button, item)
                                end
                            )
                        end
                    end
                end
            )
        end
    )
    -- 专业技能面板
    local function update_tradeskillframe_recipelist()
        for _, button in ipairs(TradeSkillFrame.RecipeList.buttons) do
            item_button_hide_all(button)
            if button and button.tradeSkillInfo and button.tradeSkillInfo.recipeID then
                local item_link = C_TradeSkillUI.GetRecipeItemLink(button.tradeSkillInfo.recipeID)
                if item_link then
                    local item = Item:CreateFromItemLink(item_link)
                    if item and not item:IsItemEmpty() then
                        item:ContinueOnItemLoad(
                            function()
                                item_button_prepare(
                                    button,
                                    function(entry)
                                        entry.mog:ClearAllPoints()
                                        entry.mog:SetPoint("LEFT", 2, 0)
                                        entry.mog:SetSize(12, 12)
                                    end
                                )
                                item_button_show_mog(button, item)
                            end
                        )
                    end
                end
            end
        end
    end
    T.On_AddOn_Loaded(
        "Blizzard_TradeSkillUI",
        function()
            -- Update on most things (like clicking)
            hooksecurefunc(TradeSkillFrame.RecipeList, "RefreshDisplay", update_tradeskillframe_recipelist)
            -- Update on scroll
            TradeSkillFrame.RecipeList.scrollBar:HookScript("OnValueChanged", update_tradeskillframe_recipelist)
            hooksecurefunc(
                TradeSkillFrame.DetailsFrame,
                "RefreshDisplay",
                function()
                    local button = TradeSkillFrame.DetailsFrame.Contents.ResultIcon
                    item_button_hide_all(button)
                    local recipe_id = TradeSkillFrame.DetailsFrame.selectedRecipeID
                    if not recipe_id then
                        return
                    end
                    local item_link = C_TradeSkillUI.GetRecipeItemLink(recipe_id)
                    if not item_link then
                        return
                    end
                    local item = Item:CreateFromItemLink(item_link)
                    if not item or item:IsItemEmpty() then
                        return
                    end
                    item:ContinueOnItemLoad(
                        function()
                            -- print(item:GetItemID())
                            item_button_prepare(
                                button,
                                function(entry)
                                    entry.overlay:SetFrameLevel(102)
                                end
                            )
                            item_button_show_ilvl(button, item)
                            item_button_show_ilvlup(button, item)
                            item_button_show_mog(button, item)
                        end
                    )
                end
            )
        end
    )
    -- 训练师面板
    T.On_AddOn_Loaded(
        "Blizzard_TrainerUI",
        function()
            hooksecurefunc(
                "ClassTrainerFrame_SetServiceButton",
                function(skill_button, skill_index, player_money, selected, is_trade_skill)
                    item_button_hide_all(skill_button)
                    local item_link = GetTrainerServiceItemLink(skill_index)
                    if item_link then
                        local item = Item:CreateFromItemLink(item_link)
                        if item and not item:IsItemEmpty() then
                            item:ContinueOnItemLoad(
                                function()
                                    item_button_prepare(
                                        skill_button,
                                        function(entry)
                                            entry.overlay:ClearAllPoints()
                                            entry.overlay:SetAllPoints(skill_button.icon)
                                            entry.overlay:SetFrameLevel(6)
                                        end
                                    )
                                    item_button_show_ilvl(skill_button, item)
                                    item_button_show_ilvlup(skill_button, item)
                                    item_button_show_mog(skill_button, item)
                                end
                            )
                        end
                    end
                end
            )
        end
    )
    -- 公会银行
    T.On_AddOn_Loaded(
        "Blizzard_GuildBankUI",
        function()
            hooksecurefunc(
                GuildBankFrame,
                "Update",
                function()
                    local tab = GetCurrentGuildBankTab()
                    for column = 1, 7 do
                        for index = 1, 14 do
                            local slot = (column - 1) * 14 + index
                            -- print(column, index, slot)
                            local item_link = GetGuildBankItemLink(tab, slot)
                            -- print(column, index, slot, item_link)
                            local button = GuildBankFrame.Columns[column].Buttons[index]
                            item_button_hide_all(button)
                            if item_link then
                                local item = Item:CreateFromItemLink(item_link)
                                if item and (not item:IsItemEmpty()) then
                                    item:ContinueOnItemLoad(
                                        function()
                                            item_button_show_ilvl(button, item)
                                            item_button_show_ilvlup(button, item)
                                            item_button_show_mog(button, item)
                                        end
                                    )
                                end
                            end
                        end
                    end
                end
            )
        end
    )
    -- 拍卖行
    T.On_AddOn_Loaded(
        "Blizzard_AuctionHouseUI",
        function()
            hooksecurefunc(
                AuctionHouseFrame.BrowseResultsFrame.ItemList,
                "RefreshScrollFrame",
                function(self)
                    for i, button in ipairs(self.ScrollFrame.buttons) do
                        item_button_hide_all(button)
                        if button.rowData and button.rowData.itemKey and button.rowData.itemKey.itemID then
                            local item = Item:CreateFromItemID(button.rowData.itemKey.itemID)
                            if item and (not item:IsItemEmpty()) then
                                item:ContinueOnItemLoad(
                                    function()
                                        item_button_prepare(
                                            button,
                                            function(entry)
                                                entry.overlay:ClearAllPoints()
                                                entry.overlay:SetAllPoints(button.cells[2])
                                                -- entry.overlay:SetFrameLevel(6)
                                                entry.mog:ClearAllPoints()
                                                entry.mog:SetPoint("RIGHT", -2, 0)
                                                entry.mog:SetSize(12, 12)
                                            end
                                        )
                                        item_button_show_mog(button, item)
                                    end
                                )
                            end
                        end
                    end
                end
            )
        end
    )
end
