local N, T = ...

function T.Setup_Shortcut()
    local frame = CreateFrame("Frame", nil, UIParent, BackdropTemplateMixin and "BackdropTemplate")
    frame:SetBackdrop(
        {
            bgFile = "Interface/DialogFrame/UI-DialogBox-Background",
            -- edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
            edgeSize = 12,
            insets = {left = 3, right = 3, top = 3, bottom = 3}
        }
    )
    frame:SetSize(20, 20)
    frame:SetPoint("BOTTOMRIGHT", MicroButtonAndBagsBar, "BOTTOMLEFT", 0, 0)
    frame:SetParent(MicroButtonAndBagsBar)

    local frame_padding = {
        horizontal = 0,
        vertical = 0
    }
    local button_spacing = {
        horizontal = 0,
        vertical = 0
    }
    local button_padding = {
        horizontal = 6,
        vertical = 4
    }
    local prev_line_first_btn = nil
    local prev_btn = nil
    local function add_to_panel(btn, next_line)
        if next_line then
            if not prev_line_first_btn then
                btn:SetPoint("TOPLEFT", frame, "TOPLEFT", frame_padding.horizontal, -frame_padding.vertical)
            else
                btn:SetPoint("TOPLEFT", prev_line_first_btn, "BOTTOMLEFT", 0, -button_spacing.vertical)
            end
            prev_line_first_btn = btn
        else
            if prev_btn then
                btn:SetPoint("TOPLEFT", prev_btn, "TOPRIGHT", button_spacing.horizontal, 0)
            end
        end
        prev_btn = btn
    end

    local function create_macro_button(parent, text, tooltip, macrotext)
        local btn = CreateFrame("Button", nil, parent, "UIPanelButtonTemplate, SecureActionButtonTemplate")
        btn:SetText(text)
        btn:SetSize(
            btn:GetTextWidth() + button_padding.horizontal * 2,
            btn:GetTextHeight() + button_padding.vertical * 2
        )
        btn:SetAttribute("type", "macro")
        btn:SetAttribute("macrotext", macrotext)
        btn:SetScript(
            "OnEnter",
            function(self)
                GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
                GameTooltip:SetText(tooltip, 1, 1, 1, 1, true)
                GameTooltip:Show()
            end
        )
        btn:SetScript(
            "OnLeave",
            function(self)
                GameTooltip:Hide()
            end
        )
        return btn
    end

    -- 切换专精按钮
    for i = 1, GetNumSpecializations() do
        local _, sname, sdescr = GetSpecializationInfo(i)
        local btn = create_macro_button(frame, sname, sdescr, "/run SetSpecialization(" .. i .. ")")
        add_to_panel(btn, i == 1)
    end

    if T.IsAuthor then
        -- 菜单按钮和右键菜单
        local menu = {}
        table.insert(menu, {text = "≡ 区域 ≡", notCheckable = true, isTitle = true, justifyH = "CENTER"})
        for _, zone in ipairs({"艾尔文森林", "暮色森林", "赤脊山", "荆棘谷海角", "翡翠林"}) do
            table.insert(
                menu,
                {
                    text = zone,
                    notCheckable = true,
                    func = function()
                        C_FriendList.SetWhoToUi(true)
                        C_FriendList.SendWho("z-" .. zone)
                    end
                }
            )
        end
        table.insert(menu, {text = "≡ 公会 ≡", notCheckable = true, isTitle = true, justifyH = "CENTER"})
        for _, zone in ipairs({"乔纳森之盾", "夜幕之刃佣兵团", "提拉加德之傲", "塔兰图拉", "黑鸦佣兵", "闪金精神病院", "狄奥尼索斯之吻", "湖畔镇", "拜拉瑟恩之刃"}) do
            table.insert(
                menu,
                {
                    text = zone,
                    notCheckable = true,
                    func = function()
                        C_FriendList.SetWhoToUi(true)
                        C_FriendList.SendWho("g-" .. zone)
                    end
                }
            )
        end
        local menuFrame = CreateFrame("Frame", "ExampleMenuFrame", UIParent, "UIDropDownMenuTemplate")
        frame:EnableMouse(true)
        frame:SetScript(
            "OnMouseUp",
            function(self, button)
                if button == "RightButton" then
                    EasyMenu(menu, menuFrame, "cursor", 0, 0, "MENU", 3)
                end
            end
        )
        -- 菜单按钮
        local btn = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
        btn:SetText("≡")
        btn:SetSize(
            btn:GetTextWidth() + button_padding.horizontal * 2,
            btn:GetTextHeight() + button_padding.vertical * 2
        )
        btn:SetScript(
            "OnClick",
            function(self, button, down)
                EasyMenu(menu, menuFrame, self, 0, 0, "MENU", 3)
            end
        )
        add_to_panel(btn, false)
    end

    -- 固定的宏按钮
    for i, line in ipairs(
        {
            {
                T.IsAuthor and
                    {
                        "CD",
                        "免费CD君",
                        "/w 羔羊公益之使-瓦里安 9\n/w 羔羊公益之命-末日行者 9\n/w 羔羊公益之忘-金色平原 9\n/w 羔羊公益之初-瓦里安 9\n/w 羔羊公益之不-瓦拉纳 9\n/w 羔羊公益之牢-金色平原 9\n/w 羔羊公益之心-冰霜之刃 9\n"
                    },
                T.IsAuthor and {"H", "转10H难度", "/p 10H\n"},
                {"重置", "重置所有副本", "/run ResetInstances()"},
                {"公会", "旧版公会界面，会员可按最后上线时间排序", "/guildroster\n"},
                -- {"老鸟", "让队友听话", "/bg 照我的做，我是老鸟我负责\n"},
                {"退组", "退出小队、团队或者等待队列", "/script C_PartyInfo.LeaveParty()\n/script LeaveBattlefield()"},
                {"登出", "关闭自动防离并登出", "/al off\n/logout\n"}
                -- {"宏", "创建/编辑宏", "/macro"},
                -- {
                --     "自杀",
                --     "帮助界面的角色卡死自杀功能，别忘了摧毁炉石",
                --     '/run if GetItemCount("6948") == 1 then message("请摧毁炉石") end\n/click HelpFrameCharacterStuckStuck'
                -- },
                -- {
                --     "防离",
                --     "防止挂机时自动登出",
                --     '/script T,F=T or 0,F or CreateFrame("frame")if X then X=nil DEFAULT_CHAT_FRAME:AddMessage("防离关闭")else DEFAULT_CHAT_FRAME:AddMessage("防离开启") X=function()local t=GetTime()if t-T>1 then StaticPopup1Button2:Click()T=t end end end F:SetScript("OnUpdate",X)'
                -- }
            }
        }
    ) do
        local next_line = true
        for j = 1, #line do -- 不能用ipairs,因为值nil会停止遍历
            local cell = line[j]
            if cell then
                local btn = create_macro_button(frame, cell[1], cell[2], cell[3])
                add_to_panel(btn, next_line)
                next_line = false
            end
        end
    end

    -- Interface\FrameXML\SecureTemplates.lua SecureActionButton_OnClick()
    -- local btn = create_macro_button(frame, "Stuck", "", "")
    -- btn:SetAttribute("type", "Stuck")
    -- btn.Stuck = Stuck
    -- add_to_panel(btn, false)
    -- hooksecurefunc("Stuck", function()print("Stuck()")end)

    -- 重新调整面板尺寸
    local top, left, right, bottom = {}, {}, {}, {}
    for _, region in ipairs({frame:GetChildren()}) do
        table.insert(top, region:GetTop())
        table.insert(left, region:GetLeft())
        table.insert(right, region:GetRight())
        table.insert(bottom, region:GetBottom())
    end
    local w, h = math.max(unpack(right)) - math.min(unpack(left)), math.max(unpack(top)) - math.min(unpack(bottom))
    frame:SetSize(w + frame_padding.horizontal * 2, h + frame_padding.vertical * 2)
end
