local N, T = ...

function T.Setup_SpamBlocker()
    local default_history_size = 0
    local default_similarity = 0
    spamblocker_history_size = spamblocker_history_size or default_history_size
    spamblocker_similarity = spamblocker_similarity or default_similarity
    local message_history = {}
    local my_guid = nil

    local function strcos(str1, str2)
        if str1 == "" and str2 == "" then
            return 1
        elseif (str1 == "" and str2 ~= "") or (str1 ~= "" and str2 == "") then
            return 0
        else
            local ch_freq = {}
            for i = 1, string.len(str1) do
                local ch = string.byte(str1, i)
                if ch_freq[ch] then
                    ch_freq[ch][1] = ch_freq[ch][1] + 1
                else
                    ch_freq[ch] = {1, 0}
                end
            end
            for i = 1, string.len(str2) do
                local ch = string.byte(str2, i)
                if ch_freq[ch] then
                    ch_freq[ch][2] = ch_freq[ch][2] + 1
                else
                    ch_freq[ch] = {0, 1}
                end
            end
            local sum_axb, sum_as, sum_bs = 0, 0, 0
            for ch, freq in pairs(ch_freq) do
                sum_axb = sum_axb + freq[1] * freq[2]
                sum_as = sum_as + freq[1] * freq[1]
                sum_bs = sum_bs + freq[2] * freq[2]
            end
            return sum_axb / (math.sqrt(sum_as) * math.sqrt(sum_bs))
        end
    end

    local function chat_filter(self, event, ...)
        -- print(event, ...)
        if
            (not spamblocker_history_size) or spamblocker_history_size < 1 or (not spamblocker_similarity) or
                spamblocker_similarity < 0 or
                spamblocker_similarity > 1
         then
            return false
        end
        local guid = select(12, ...)
        if (not guid) or guid == my_guid then
            return false
        end
        local msg = select(1, ...)
        if not msg then
            return false
        end
        local name = select(2, ...)
        if not name then
            return false
        end
        if not message_history[guid] then
            message_history[guid] = {
                index = 0,
                cache = {}
            }
        end
        local history = message_history[guid]
        local matched = false
        if event == "CHAT_MSG_WHISPER" and IsInInstance() then
            matched = true
        else
            for i = 1, #history.cache do
                if history.cache[i] and strcos(history.cache[i], msg) >= spamblocker_similarity then
                    matched = true
                    break
                end
            end
        end
        history.index = history.index + 1
        if history.index > spamblocker_history_size then
            history.index = 1
        end
        history.cache[history.index] = msg
        if matched then
            -- DEFAULT_CHAT_FRAME:AddMessage("|cFF666666[" .. name .. "] " .. msg)
            if event == "CHAT_MSG_SAY" or event == "CHAT_MSG_YELL" then
                SetCVar("chatBubbles", 0)
            end
            return true
        else
            SetCVar("chatBubbles", 1)
            return false
        end
    end

    my_guid = UnitGUID("player")
    for _, event in ipairs(
        {
            "CHAT_MSG_SAY",
            "CHAT_MSG_YELL",
            "CHAT_MSG_WHISPER",
            "CHAT_MSG_TEXT_EMOTE",
            "CHAT_MSG_EMOTE",
            "CHAT_MSG_DND",
            "CHAT_MSG_AFK"
            -- "CHAT_MSG_CHANNEL", -- 有问题，同一条消息会在chat_filter重复数十次
            -- "CHAT_MSG_TRADESKILLS" -- guid都是nil
        }
    ) do
        ChatFrame_AddMessageEventFilter(event, chat_filter)
    end
    local function print_spamblocker_state()
        print(" ")
        print("spamblocker_history_size=", spamblocker_history_size)
        print("spamblocker_similarity=", spamblocker_similarity)
    end
    T.Add_Slash_Command(
        {
            name = "SPAMBLOCKER_CONFIG",
            commands = {"/sb"},
            callback = function(...)
                local args = {...}
                if args[1] == "set" then
                    local hsize = math.floor(tonumber(args[2]))
                    if hsize then
                        spamblocker_history_size = hsize
                    end
                    local simi = tonumber(args[3])
                    if simi then
                        spamblocker_similarity = simi
                    end
                    message_history = {}
                elseif args[1] == "reset" then
                    spamblocker_history_size = default_history_size
                    spamblocker_similarity = default_similarity
                    message_history = {}
                end
                print_spamblocker_state()
            end
        }
    )
end
